package de.fhdw.gaming.ipspiel23.memory;

import de.fhdw.gaming.core.domain.Move;
import de.fhdw.gaming.core.domain.Player;
import de.fhdw.gaming.core.domain.State;
import de.fhdw.gaming.core.domain.Strategy;

/**
 * A handle uniquely identifying a game memory instance.
 */
public final class GameMemoryIdentifier implements IGameMemoryIdentifier {

    /**
     * The hash code/identity value of this identifier.
     */
    private final int precomputedHashCode;

    /**
     * Creates a new identifier with the given hash code.
     *
     * @param hashCode The hash code/identity value of this identifier.
     */
    private GameMemoryIdentifier(final int hashCode) {
        precomputedHashCode = hashCode;
    }

    /**
     * Creates a new identifier for the given player and strategy against the given opponent strategy.
     * Subsequent calls with the same parameters will return the same identifier instance.
     * 
     * @param <P> The player type.
     * @param <S> The state type.
     * @param <M> The move type.
     * @param player The player.
     * @param strategy The strategy.
     * @param opponentStrategy The opponent's strategy.
     * @return A new identifier for the given player and strategy against the given opponent strategy.
     */
    public static <P extends Player<P>, S extends State<P, S>, M extends Move<P, S>> 
        IGameMemoryIdentifier of(final Player<P> player, final Strategy<P, S, M> strategy, 
            final Strategy<P, S, M> opponentStrategy) {
        
        // pre-compute the hash code to avoid having to store this generic madness.
        // reuse identifier is based on strategy type, and player name.
        int hash = 7;
        hash = 31 * hash + player.getName().hashCode();
        hash = 31 * hash + strategy.getClass().getName().hashCode();
        hash = 31 * hash + opponentStrategy.getClass().getName().hashCode();
        return new GameMemoryIdentifier(hash);
    }

    @Override
    public int hashCode() {
        return precomputedHashCode;
    }

    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof GameMemoryIdentifier) {
            return precomputedHashCode == ((GameMemoryIdentifier) obj).precomputedHashCode;
        }
        return false;
    }
}
